<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace MapModule\Model\Table;

use Cake\Datasource\EntityInterface;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use MapModule\Model\Entity\Mapitem;

/**
 * Mapitems Model
 *
 * @property MapsTable&BelongsTo $Maps
 *
 * @method Mapitem get($primaryKey, $options = [])
 * @method Mapitem newEntity($data = null, array $options = [])
 * @method Mapitem[] newEntities(array $data, array $options = [])
 * @method Mapitem|false save(EntityInterface $entity, $options = [])
 * @method Mapitem saveOrFail(EntityInterface $entity, $options = [])
 * @method Mapitem patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Mapitem[] patchEntities($entities, array $data, array $options = [])
 * @method Mapitem findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin TimestampBehavior
 */
class MapitemsTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('mapitems');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Maps', [
            'foreignKey' => 'map_id',
            'joinType'   => 'INNER',
            'className'  => 'MapModule.Maps',
        ]);

        $this->hasMany('Hosts');
        $this->hasMany('Hostgroups');
        $this->hasMany('Services');
        $this->hasMany('Servicegroups');
    }

    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosts':
                $coreTable->hasMany('Mapitems', [
                    'className'  => 'MapModule.Mapitems',
                    'dependent'  => true,
                    'foreignKey' => 'object_id',
                    'joinType'   => 'INNER',
                    'conditions' => [
                        'type' => 'host'
                    ]
                ]);
                break;
            case 'Hostgroups':
                $coreTable->hasMany('Mapitems', [
                    'className'  => 'MapModule.Mapitems',
                    'dependent'  => true,
                    'foreignKey' => 'object_id',
                    'joinType'   => 'INNER',
                    'conditions' => [
                        'type' => 'hostgroup'
                    ]
                ]);
                break;
            case 'Services':
                $coreTable->hasMany('Mapitems', [
                    'className'  => 'MapModule.Mapitems',
                    'dependent'  => true,
                    'foreignKey' => 'object_id',
                    'joinType'   => 'INNER',
                    'conditions' => [
                        'type' => 'service'
                    ]
                ]);
                break;
            case 'Servicegroups':
                $coreTable->hasMany('Mapitems', [
                    'className'  => 'MapModule.Mapitems',
                    'dependent'  => true,
                    'foreignKey' => 'object_id',
                    'joinType'   => 'INNER',
                    'conditions' => [
                        'type' => 'servicegroup'
                    ]
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn(['map_id'], 'Maps'));

        return $rules;
    }

    /**
     * @param $objectId
     * @param $mapId
     * @return array
     */
    public function getMapitemsForMaps($objectId, $mapId) {
        $query = $this->find()
            ->select(['Mapitems.object_id'])
            ->where([
                'Mapitems.map_id' => $mapId,
                'Mapitems.type'   => 'map',
            ]);
        if ($objectId !== null) {
            $query->where([
                'Mapitems.object_id' => $objectId,
            ]);
        }

        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $result->toArray();
    }

    /**
     * @param $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Mapitems.id' => $id]);
    }

    /**
     * @param int $mapId
     * @param bool $enableHydration
     * @return array
     */
    public function allVisibleMapItems($mapId, $enableHydration = true) {
        $query = $this->find()
            ->join([
                [
                    'table'      => 'maps',
                    'type'       => 'INNER',
                    'alias'      => 'Maps',
                    'conditions' => 'Maps.id = Mapitems.map_id',
                ]
            ])
            ->select([
                'Mapitems.map_id',
                'Mapitems.object_id',
            ])
            ->where([
                'Mapitems.type'       => 'map',
                'Mapitems.map_id != ' => $mapId
            ])
            ->enableHydration($enableHydration);

        $result = $query->all();
        if (empty($result)) {
            return [];
        }
        return $result->toArray();
    }

}
