<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace App\Lib;

use Acl\Model\Table\AcosTable;

/**
 * Class AclDependencies
 * @package App
 */
class AclDependencies {

    /**
     * Hold a list of controllers and actions which
     * should always be allowed and could not be disabled by the user
     *
     * For Example:
     * PagesController::paginator()
     * PagesController::csrf()
     * Users::login()
     * Users::logout()
     *
     * @var array
     */
    private $allow = [];

    /**
     * Controller actions that depends on other controller actions.
     * This often happens when using a lot of Ajax in your frontend.
     *
     * For example:
     * Users::edit() make an Ajax call to Users::loadUsergroups() etc...
     *
     * @var array
     */
    private $dependencies = [];

    /**
     * Action names that should be ignored and not displayed by the frontend
     * @var array
     */
    private $ignore = [
        'initializePluginTables'
    ];

    /**
     * AclDependencies constructor.
     */
    public function __construct() {

        // Add actions that should always be allowed.
        $this
            ->allow('Angular', 'sidebar')
            ->allow('Angular', 'menu')
            ->allow('Angular', 'user_timezone')
            ->allow('Angular', 'menustats')
            ->allow('Angular', 'paginator')
            ->allow('Angular', 'scroll')
            ->allow('Angular', 'mass_delete')
            ->allow('Angular', 'topSearch')
            ->allow('Angular', 'not_found')
            ->allow('Angular', 'forbidden')
            ->allow('Angular', 'reload_required')
            ->allow('Angular', 'reschedule_host')
            ->allow('Angular', 'downtime_host')
            ->allow('Angular', 'getDowntimeData')
            ->allow('Angular', 'acknowledge_host')
            ->allow('Angular', 'submit_host_result')
            ->allow('Angular', 'disable_host_flap_detection')
            ->allow('Angular', 'enable_host_flap_detection')
            ->allow('Angular', 'executing')
            ->allow('Angular', 'downtime_service')
            ->allow('Angular', 'acknowledge_service')
            ->allow('Angular', 'submit_service_result')
            ->allow('Angular', 'disable_service_flap_detection')
            ->allow('Angular', 'enable_service_flap_detection')
            ->allow('Angular', 'mass_delete_host_downtimes')
            ->allow('Angular', 'mass_delete_service_downtimes')
            ->allow('Angular', 'popover_graph')
            ->allow('Angular', 'getHostAndServiceStateSummaryIcon');

        $this
            ->allow('Pages', 'index')
            ->allow('Pages', 'configuration');

        $this
            ->allow('Dashboard', 'index');

        $this
            ->allow('Users', 'login')
            ->allow('Users', 'logout');

        $this
            ->allow('Hosts', 'icon')
            ->allow('Hosts', 'menu');

        $this
            ->allow('Services', 'icon')
            ->allow('Services', 'menu');

        $this
            ->allow('Profile', 'edit')
            ->allow('Profile', 'changePassword')
            ->allow('Profile', 'upload_profile_icon')
            ->allow('Profile', 'deleteImage')
            ->allow('Profile', 'apikey')
            ->allow('Profile', 'edit_apikey')
            ->allow('Profile', 'delete_apikey')
            ->allow('Profile', 'create_apikey');

        $this
            ->allow('Downtimes', 'validateDowntimeInputFromAngular');

        $this
            ->allow('Users', 'loadDateformats');

        $this
            ->allow('Graphgenerators', 'getPerfdataByUuid');

        ///////////////////////////////
        //    Add dependencies       //
        //////////////////////////////


        $this
            ->dependency('Services', 'automap', 'Hosts', 'hoststatusdetailsmodal')
            ->dependency('Services', 'automap', 'Services', 'servicestatusdetailsmodal')
            ->dependency('Services', 'automap', 'Angular', 'hosticon')
            ->dependency('Services', 'automap', 'Angular', 'serviceicon');

        $this
            ->dependency('Users', 'add', 'Users', 'loadUsergroups')
            ->dependency('Users', 'edit', 'Users', 'loadUsergroups');

        $this
            ->dependency('Agentconnector', 'overview', 'Agentconnector', 'pull')
            ->dependency('Agentconnector', 'overview', 'Agentconnector', 'push');

        //Load Plugin ALC Dependencies
        foreach (PluginManager::getAvailablePlugins() as $pluginName) {
            $className = sprintf('\%s\Lib\AclDependencies', $pluginName);
            if (class_exists($className)) {
                /** @var PluginAclDependencies $PluginAclDependencies */
                $PluginAclDependencies = new $className();

                $this->allow[$pluginName] = $PluginAclDependencies->getAllow();
                $this->dependencies[$pluginName] = $PluginAclDependencies->getDependencies();
            }
        }
    }

    /**
     * @param string $controller
     * @param string $action
     * @return $this
     */
    public function allow(string $controller, string $action): self {
        if (!isset($this->allow[$controller])) {
            $this->allow[$controller] = [];
        }

        $this->allow[$controller][$action] = $action;
        return $this;
    }

    /**
     * @param string $controller
     * @param string $action
     * @param string $dependentController
     * @param $dependentAction
     * @return $this
     */
    public function dependency(string $controller, string $action, string $dependentController, $dependentAction): self {
        if (!isset($this->dependencies[$controller][$action])) {
            $this->dependencies[$controller][$action] = [];
        }

        if (!isset($this->dependencies[$controller][$action][$dependentController])) {
            $this->dependencies[$controller][$action][$dependentController] = [];
        }

        $this->dependencies[$controller][$action][$dependentController][] = $dependentAction;


        return $this;
    }

    /**
     * @param AcosTable $AcosTable
     * @param array $selectedAcos
     * @return array
     */
    public function getDependentAcos(AcosTable $AcosTable, array $selectedAcos): array {
        $threadedAcos = $AcosTable->find('threaded')
            ->disableHydration()
            ->all();

        $acos = [];
        foreach ($threadedAcos as $threadedAco) {
            foreach ($threadedAco['children'] as $controllerAcos) {
                if (substr($controllerAcos['alias'], -6) === 'Module') {
                    $pluginAcos = $controllerAcos;
                    foreach ($pluginAcos['children'] as $pluginControllerAcos) {
                        $acos[$pluginAcos['alias']][$pluginControllerAcos['alias']] = [
                            'id'      => $pluginControllerAcos['id'],
                            'actions' => []
                        ];
                        foreach ($pluginControllerAcos['children'] as $pluginActionAcos) {
                            $acos[$pluginAcos['alias']][$pluginControllerAcos['alias']]['actions'][$pluginActionAcos['alias']] = [
                                'id' => $pluginActionAcos['id'],
                            ];
                        }
                    }
                } else {
                    //Core
                    $acos[$controllerAcos['alias']] = [
                        'id'      => $controllerAcos['id'],
                        'actions' => []
                    ];
                    foreach ($controllerAcos['children'] as $actionAco) {
                        $acos[$controllerAcos['alias']]['actions'][$actionAco['alias']] = [
                            'id' => $actionAco['id'],
                        ];
                    }
                }
            }
        }

        //Add always allowed ACL actions to $selectedAcos
        foreach ($this->allow as $controller => $actions) {
            if (substr($controller, -6) === 'Module') {
                $pluginName = $controller;
                foreach ($actions as $pluginController => $pluginActions) {
                    foreach ($pluginActions as $pluginAction) {
                        if (isset($acos[$pluginName][$pluginController]['actions'][$pluginAction]['id'])) {
                            //debug(implode('/', [$pluginName, $pluginController, $pluginAction]));
                            $acoId = $acos[$pluginName][$pluginController]['actions'][$pluginAction]['id'];

                            $selectedAcos[$acoId] = 1;
                        }
                    }
                }
            } else {
                //Core
                foreach ($actions as $action) {
                    if (isset($acos[$controller]['actions'][$action]['id'])) {
                        $acoId = $acos[$controller]['actions'][$action]['id'];

                        $selectedAcos[$acoId] = 1;
                    }
                }
            }
        }

        //Build up dependency tree (dependent ACL actions)
        $dependencyTree = [];
        foreach ($this->dependencies as $controller => $actions) { //1 core
            if (substr($controller, -6) === 'Module') {
                $pluginName = $controller;
                $pluginControllers = $actions;
                foreach ($pluginControllers as $pluginController => $pluginControllerActions) { // like 1 for core
                    foreach ($pluginControllerActions as $pluginAction => $dependentPluginControllers) { //like 2 for core
                        foreach ($dependentPluginControllers as $dependentPluginController => $dependentPluginActions) { // like 3 for core
                            foreach ($dependentPluginActions as $dependentPluginAction) { // like 4 for core
                                //debug(sprintf(
                                //    '[%s] %s/%s depends on %s/%s',
                                //    $pluginName,
                                //    $pluginController,
                                //    $pluginAction,
                                //    $dependentPluginController,
                                //    $dependentPluginAction
                                //));

                                if (isset($acos[$pluginName][$pluginController]['actions'][$pluginAction]['id'])) {
                                    if (isset($acos[$pluginName][$dependentPluginController]['actions'][$dependentPluginAction]['id'])) {
                                        $acoId = $acos[$pluginName][$pluginController]['actions'][$pluginAction]['id'];
                                        $dependentAcoId = $acos[$pluginName][$dependentPluginController]['actions'][$dependentPluginAction]['id'];

                                        $dependencyTree[$acoId][] = $dependentAcoId;
                                    }
                                }
                            }
                        }
                    }
                }
            } else {
                //Core
                foreach ($actions as $action => $dependentControllers) { //2 core
                    foreach ($dependentControllers as $dependentController => $dependentActions) { //3 core
                        foreach ($dependentActions as $dependentAction) { //4 core
                            //debug(sprintf(
                            //    '%s/%s depends on %s/%s',
                            //    $controller,
                            //    $action,
                            //    $dependentController,
                            //    $dependentAction
                            //));

                            if (isset($acos[$controller]['actions'][$action]['id'])) {
                                if (isset($acos[$dependentController]['actions'][$dependentAction]['id'])) {
                                    $acoId = $acos[$controller]['actions'][$action]['id'];
                                    $dependentAcoId = $acos[$dependentController]['actions'][$dependentAction]['id'];

                                    $dependencyTree[$acoId][] = $dependentAcoId;
                                }
                            }
                        }
                    }
                }
            }
        }

        //Add dependent ACL actions to $selectedAcos
        foreach ($selectedAcos as $acoId => $permissions) {
            if ($permissions === 1) {
                if (isset($dependencyTree[$acoId])) {
                    //Add dependencies to $selectedAcos;

                    foreach ($dependencyTree[$acoId] as $dependentAcoId) {
                        $selectedAcos[$dependentAcoId] = 1;
                    }
                }
            }
        }

        return $selectedAcos;
    }

    public function filterAcosForFrontend($acosResultThreaded) {
        $allDependenciesSimplified = [];
        foreach ($this->dependencies as $controllerName => $actions) { //1
            if (substr($controllerName, -6) === 'Module') {
                $pluginName = $controllerName;
                $pluginControllers = $actions;

                foreach ($pluginControllers as $pluginControllerName => $pluginActions) { //Like 1 for core
                    foreach ($pluginActions as $pluginActionName => $dependentPluginController) { //Like 2 for core
                        foreach ($dependentPluginController as $dependentPluginControllerName => $dependentPluginActions) { //Like 3 for core
                            foreach ($dependentPluginActions as $dependentPluginAction) { //Like 4 for core
                                if (!isset($allDependenciesSimplified[$pluginName][$dependentPluginControllerName])) {
                                    $allDependenciesSimplified[$pluginName][$dependentPluginControllerName] = [];
                                }

                                $allDependenciesSimplified[$pluginName][$dependentPluginControllerName][$dependentPluginAction] = $dependentPluginAction;
                            }
                        }
                    }
                }
            } else {
                //Core
                foreach ($actions as $actionName => $dependentController) { //2
                    foreach ($dependentController as $dependentControllerName => $dependentActions) { //3
                        foreach ($dependentActions as $dependentAction) { //4
                            if (!isset($allDependenciesSimplified[$dependentControllerName])) {
                                $allDependenciesSimplified[$dependentControllerName] = [];
                            }

                            $allDependenciesSimplified[$dependentControllerName][$dependentAction] = $dependentAction;
                        }
                    }
                }
            }
        }

        foreach ($acosResultThreaded[0]['children'] as $controllerIndex => $controller) {
            if (substr($controller['alias'], -6) === 'Module') {
                $pluginName = $controller['alias'];
                foreach ($controller['children'] as $pluginControllerIndex => $pluginController) {
                    $pluginControllerName = $pluginController['alias'];
                    foreach ($pluginController['children'] as $pluginActionIndex => $pluginAction) {
                        $pluginActionName = $pluginAction['alias'];

                        // Remove ACOs that are always allow (the user cannot untick them!)
                        if (isset($this->allow[$pluginName][$pluginControllerName][$pluginActionName])) {
                            unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$pluginControllerIndex]['children'][$pluginActionIndex]);
                        }

                        // Remove ACOs if they are dependencies of other ACOs (the user cannot untick them!)
                        if (isset($allDependenciesSimplified[$pluginName][$pluginControllerName][$pluginActionName])) {
                            unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$pluginControllerIndex]['children'][$pluginActionIndex]);
                        }

                        // Remove ACOs that shoud be ignored like public functions form AppController that exists in all controllers due to class FooController extends AppController and so on
                        if (in_array($pluginActionName, $this->ignore, true)) {
                            unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$pluginControllerIndex]['children'][$pluginActionIndex]);
                        }
                    }
                }

            } else {
                //Core
                $controllerName = $controller['alias'];
                foreach ($controller['children'] as $actionIndex => $action) {
                    $actionName = $action['alias'];

                    // Remove ACOs that are always allow (the user cannot untick them!)
                    if (isset($this->allow[$controllerName][$actionName])) {
                        unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$actionIndex]);
                    }

                    // Remove ACOs if they are dependencies of other ACOs (the user cannot untick them!)
                    if (isset($allDependenciesSimplified[$controllerName][$actionName])) {
                        unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$actionIndex]);
                    }

                    // Remove ACOs that shoud be ignored like public functions form AppController that exists in all controllers due to class FooController extends AppController and so on
                    if (in_array($actionName, $this->ignore, true)) {
                        unset($acosResultThreaded[0]['children'][$controllerIndex]['children'][$actionIndex]);
                    }
                }
            }

            //Make sure we have arrays [] not hasmaps {} !!!
            $acosResultThreaded[0]['children'][$controllerIndex]['children'] = array_values($acosResultThreaded[0]['children'][$controllerIndex]['children']);

        }

        return $acosResultThreaded;
    }

}

