<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace App\Model\Table;

use App\Filter\GenericFilter;
use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\PushAgent;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\FileDebugger;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * PushAgents Model
 **
 * @method \App\Model\Entity\PushAgent newEmptyEntity()
 * @method \App\Model\Entity\PushAgent newEntity(array $data, array $options = [])
 * @method \App\Model\Entity\PushAgent[] newEntities(array $data, array $options = [])
 * @method \App\Model\Entity\PushAgent get($primaryKey, $options = [])
 * @method \App\Model\Entity\PushAgent findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \App\Model\Entity\PushAgent patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \App\Model\Entity\PushAgent[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \App\Model\Entity\PushAgent|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \App\Model\Entity\PushAgent saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \App\Model\Entity\PushAgent[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \App\Model\Entity\PushAgent[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \App\Model\Entity\PushAgent[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \App\Model\Entity\PushAgent[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class PushAgentsTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('push_agents');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->hasOne('HostMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'joinType'   => 'INNER'
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->scalar('uuid')
            ->maxLength('uuid', 37)
            ->requirePresence('uuid', 'create')
            ->notEmptyString('uuid')
            ->add('uuid', 'unique', ['rule' => 'validateUnique', 'provider' => 'table']);

        $validator
            ->scalar('host_uuid')
            ->maxLength('host_uuid', 37)
            ->allowEmptyString('host_uuid');

        $validator
            ->scalar('password')
            ->maxLength('password', 255)
            ->requirePresence('password', 'create')
            ->notEmptyString('password');

        $validator
            ->scalar('hostname')
            ->maxLength('hostname', 255)
            ->allowEmptyString('hostname');

        $validator
            ->scalar('ipaddress')
            ->maxLength('ipaddress', 255)
            ->allowEmptyString('ipaddress');

        $validator
            ->scalar('remote_address')
            ->maxLength('remote_address', 255)
            ->allowEmptyString('remote_address');

        $validator
            ->scalar('http_x_forwarded_for')
            ->maxLength('http_x_forwarded_for', 255)
            ->allowEmptyString('http_x_forwarded_for');

        $validator
            ->scalar('checkresults')
            ->maxLength('checkresults', 16777215)
            ->allowEmptyString('checkresults');

        $validator
            ->dateTime('last_update')
            ->notEmptyDateTime('last_update');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->isUnique(['uuid']), ['errorField' => 'uuid']);
        return $rules;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['PushAgents.id' => $id]);
    }

    /**
     * @param string $uuid
     * @return bool
     */
    public function existsByUuid($uuid) {
        return $this->exists(['PushAgents.uuid' => $uuid]);
    }

    /**
     * @param string $uuid
     * @param string $password
     * @return bool
     */
    public function existsByUuidAndPassword($uuid, $password) {
        return $this->exists([
            'PushAgents.uuid'     => $uuid,
            'PushAgents.password' => $password
        ]);
    }

    /**
     * @param string $hostuuid
     * @return bool
     */
    public function existsByHostUuidId($hostuuid) {
        return $this->exists([
            'PushAgents.host_uuid' => $hostuuid
        ]);
    }

    /**
     * @param string $uuid
     * @param string $password
     * @return PushAgent|EntityInterface
     * @throws RecordNotFoundException
     */
    public function getConfigWithHostForSubmitCheckdata($uuid, $password) {
        $query = $this->find()
            ->where([
                'PushAgents.uuid'     => $uuid,
                'PushAgents.password' => $password,
                'PushAgents.host_uuid IS NOT NULL'
            ]);
        return $query->firstOrFail();
    }

    /**
     * @param string $uuid
     * @param string $password
     * @return PushAgent|EntityInterface
     * @throws RecordNotFoundException
     */
    public function getRecordByAgentUuidAndPassword($uuid, $password) {
        $query = $this->find()
            ->where([
                'PushAgents.uuid'     => $uuid,
                'PushAgents.password' => $password
            ]);
        return $query->firstOrFail();
    }

    /**
     * @param string|null $hostuuid
     * @return array
     */
    public function getPushAgentsForAssignments(?string $hostuuid = null) {
        $query = $this->find();
        if ($hostuuid !== null) {
            $query->where([
                'OR' => [
                    'PushAgents.host_uuid' => $hostuuid,
                    'PushAgents.host_uuid IS NULL'
                ]
            ]);
        } else {
            $query->where('PushAgents.host_uuid IS NULL');
        }
        $query->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }

    /**
     * @return array
     */
    public function getAllPushAgents() {
        $query = $this->find();
        $query->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }

    /**
     * @param string $uuid
     * @return array
     */
    public function getPushAgentsByUuid(string $uuid) {
        return $this->find()
            ->where([
                'PushAgents.uuid' => $uuid,
            ])->first();
    }

    /**
     * @param $hostUuid
     * @return array|EntityInterface|null
     */
    public function getPushAgentForFreshnessCheck($hostUuid) {
        return $this->find()
            ->where([
                'PushAgents.host_uuid' => $hostUuid,
            ])->first();
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getPushAgents(GenericFilter $GenericFilter, PaginateOMat $PaginateOMat = null) {
        // Yes - this query is from hell!

        $query = $this->find();
        $query->select([
            'PushAgents.id',
            'PushAgents.uuid',
            'PushAgents.host_uuid',
            'PushAgents.hostname',
            'HostMapping.uuid',
            'HostMapping.display_name',
            'PushAgents.ipaddress',
            'PushAgents.remote_address',
            'PushAgents.http_x_forwarded_for',
            //'PushAgents.checkresults', // The Agent check result is not necessary, slows down the query and uses A LOT of memory with large paginator limits
            'PushAgents.last_update'
        ])
            ->leftJoin(
                ['HostMapping' => 'mapping'],
                ['PushAgents.host_uuid = HostMapping.uuid', 'HostMapping.type_id' => 256]
            );

        $where = $GenericFilter->genericFilters();
        if (isset($where['hasHostAssignment'])) {
            if ($where['hasHostAssignment'] == 1) {
                $query->whereNotNull([
                    'HostMapping.uuid'
                ]);
            } else {
                $query->whereNull([
                    'HostMapping.uuid'
                ]);
            }
            unset($where['hasHostAssignment']);
        }
        $query->where($where);
        $query->disableHydration();
        $query->order(
            array_merge(
                $GenericFilter->getOrderForPaginator('HostMapping.display_name', 'asc'),
                ['PushAgents.id' => 'asc']
            )

        );
        $query->group('PushAgents.id');

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }
}
