<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace App\Model\Table;

use App\Filter\ServicestatusFilter;
use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\Servicestatus;
use Cake\Datasource\EntityInterface;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Servicestatus Model
 *
 * @method Servicestatus get($primaryKey, $options = [])
 * @method Servicestatus newEntity($data = null, array $options = [])
 * @method Servicestatus[] newEntities(array $data, array $options = [])
 * @method Servicestatus|bool save(EntityInterface $entity, $options = [])
 * @method Servicestatus patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Servicestatus[] patchEntities($entities, array $data, array $options = [])
 * @method Servicestatus findOrCreate($search, callable $callback = null, $options = [])
 */
class ServicestatusTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('statusengine_servicestatus');
        $this->setDisplayField('hostname');
        $this->setPrimaryKey(['service_description']);

        $this->hasOne('ServiceMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'joinType'   => 'INNER'
        ]);

        $this->hasOne('HostMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'bindingKey' => 'hostname',
            'joinType'   => 'INNER'
        ]);

        $this->hasOne('TimeperiodMapping', [
            'className'  => 'Mapping',
            'foreignKey' => 'uuid',
            'bindingKey' => 'check_timeperiod',
            'joinType'   => 'INNER'
        ]);

        $this->hasOne('Hoststatus', [
            'className'  => 'Hoststatus',
            'foreignKey' => 'hostname',
            'bindingKey' => 'hostname',
            'joinType'   => 'INNER'
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        //Readonly table
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        //Readonly table
        return $rules;
    }

    /**
     * @param bool $includeOkState
     * @return array
     */
    public function getServicestatusCountStatusengine3($includeOkState = false) {
        $servicestatusCount = [
            '1' => 0,
            '2' => 0,
            '3' => 0,
        ];
        if ($includeOkState === true) {
            $servicestatusCount['0'] = 0;
        }

        $query = $this->find();
        $query
            ->select([
                'current_state',
                'count' => $query->newExpr('COUNT(current_state)'),
            ])
            ->group([
                'current_state',
            ])
            ->disableHydration();

        if ($includeOkState === false) {
            $query->andWhere([
                'current_state >' => 0
            ]);
        }

        $servicestatusCountResult = $query->all();

        foreach ($servicestatusCountResult as $servicestatus) {
            $servicestatusCount[$servicestatus['current_state']] = (int)$servicestatus['count'];
        }
        return $servicestatusCount;
    }


    /**
     * @param ServicestatusFilter $ServicestatusFilter
     * @param PaginateOMat|null $PaginateOMat
     */
    public function getServicestatusIndex(ServicestatusFilter $ServicestatusFilter, ?PaginateOMat $PaginateOMat = null) {
        $query = $this->find();
        $query->contain([
            'ServiceMapping',
            'HostMapping',
            'Hoststatus'
        ]);

        $query->where($ServicestatusFilter->indexFilter());
        $query->order($ServicestatusFilter->getOrderForPaginator('Servicestatus.current_state', 'desc'));


        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }

    /**
     * @param string $hostname
     * @return array
     */
    public function getSummaryStateForHoststatusIndex(string $hostname) {
        $query = $this->find();
        $query->select([
            'current_state',
            'count' => $query->func()->count('*')
        ])
            ->where(['hostname' => $hostname])
            ->group('current_state')
            ->disableHydration();

        $serviceSummaryState = [
            'ok'       => 0,
            'warning'  => 0,
            'critical' => 0,
            'unknown'  => 0
        ];

        $result = $query->all();
        if ($result === null) {
            return $serviceSummaryState;
        }

        $Servicestatus = new \App\Servicestatus();
        foreach ($result->toArray() as $record) {
            $serviceSummaryState[$Servicestatus->getStatenameByStateId($record['current_state'])] = $record['count'];
        }
        return $serviceSummaryState;
    }

    /**
     * @param string $uuid
     * @return Servicestatus|null
     */
    public function getServicestatusDetail(string $uuid): ?Servicestatus {
        $serviceStatus = $this->find()
            ->contain([
                'ServiceMapping',
                'HostMapping',
                'TimeperiodMapping'
            ])
            ->where(['service_description' => $uuid])
            ->first();

        return $serviceStatus;
    }

    /**
     * @return int
     */
    public function getServicestatusCountForDashboard() {
        return $this->find()->count();
    }

    /**
     * @return array
     */
    public function getServicestateSummary() {
        $serviceSummaryState = [
            'ok'       => 0,
            'warning'  => 0,
            'critical' => 0,
            'unknown'  => 0
        ];

        $query = $this->find();
        $query->select([
            'current_state',
            'count' => $query->func()->count('*')
        ])
            ->group('current_state');

        $Servicestatus = new \App\Servicestatus();
        foreach ($query as $record) {
            $serviceSummaryState[$Servicestatus->getStatenameByStateId($record->current_state)] = $record->count;
        }

        return $serviceSummaryState;
    }

    /**
     * @param ServicestatusFilter $ServicestatusFilter
     * @return array
     */
    public function getServicesForAutomap(ServicestatusFilter $ServicestatusFilter) {
        $query = $this->find();
        $query->select([
            'Servicestatus.hostname',
            'Servicestatus.service_description',
            'Servicestatus.current_state',
            'Servicestatus.active_checks_enabled',
            'Servicestatus.last_state_change',
            'Servicestatus.last_check',
            'Servicestatus.next_check',
            'Servicestatus.output',
            'Servicestatus.is_flapping',
            'Servicestatus.problem_has_been_acknowledged',
            'Servicestatus.scheduled_downtime_depth',

            'HostMapping.display_name',

            'ServiceMapping.display_name',

            'Hoststatus.current_state',
            'Hoststatus.is_flapping',
            'Hoststatus.problem_has_been_acknowledged',
            'Hoststatus.scheduled_downtime_depth',
            'Hoststatus.flap_detection_enabled'
        ])
            ->join([
                'ServiceMapping' => [
                    'table'      => 'mapping',
                    'alias'      => 'ServiceMapping',
                    'type'       => 'INNER',
                    'foreignKey' => false,
                    'conditions' =>
                        'ServiceMapping.uuid = Servicestatus.service_description'

                ],
                'HostMapping'    => [
                    'table'      => 'mapping',
                    'alias'      => 'HostMapping',
                    'foreignKey' => false,
                    'type'       => 'INNER',
                    'conditions' => 'HostMapping.uuid = Servicestatus.hostname'

                ],
                'Hoststatus'     => [
                    'table'      => 'statusengine_hoststatus',
                    'alias'      => 'Hoststatus',
                    'foreignKey' => false,
                    'type'       => 'INNER',
                    'conditions' => 'Hoststatus.hostname = Servicestatus.hostname'

                ]
            ])
            ->where($ServicestatusFilter->indexFilter())
            ->order([
                'HostMapping.display_name'    => 'asc',
                'ServiceMapping.display_name' => 'asc'
            ])
            ->disableHydration();

        $services = $query->all();

        if ($services === null) {
            return [];
        }

        return $services->toArray();
    }


    /**
     * @param $servicestatus
     * @param bool $extended show details ('acknowledged', 'in downtime', ...)
     * @return array
     */
    public function getServiceStateSummaryFromServicestatus($servicestatus, $extended = true) {
        $serviceStateSummary = [
            'state' => [
                0 => 0,
                1 => 0,
                2 => 0,
                3 => 0
            ],
            'total' => 0
        ];
        if ($extended === true) {
            $serviceStateSummary = [
                'state'        => [
                    0 => 0,
                    1 => 0,
                    2 => 0,
                    3 => 0
                ],
                'acknowledged' => [
                    0 => 0,
                    1 => 0,
                    2 => 0,
                    3 => 0
                ],
                'in_downtime'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0,
                    3 => 0
                ],
                'not_handled'  => [
                    0 => 0,
                    1 => 0,
                    2 => 0,
                    3 => 0
                ],
                'passive'      => [
                    0 => 0,
                    1 => 0,
                    2 => 0,
                    3 => 0
                ],
                'total'        => 0
            ];
        }
        if (empty($servicestatus)) {
            return $serviceStateSummary;
        }

        foreach ($servicestatus as $serviceUuid => $status) {
            //Check for randome exit codes like 255...
            if ($status['current_state'] > 3) {
                $status['current_state'] = 3;
            }

            $serviceStateSummary['state'][$status['current_state']]++;
            if ($extended === true) {
                if ($status['current_state'] > 0) {
                    if ($status['problem_has_been_acknowledged'] > 0) {
                        $serviceStateSummary['acknowledged'][$status['current_state']]++;
                    } else {
                        $serviceStateSummary['not_handled'][$status['current_state']]++;
                    }
                }

                if ($status['scheduled_downtime_depth'] > 0) {
                    $serviceStateSummary['in_downtime'][$status['current_state']]++;
                }
                if ($status['active_checks_enabled'] == 0) {
                    $serviceStateSummary['passive'][$status['current_state']]++;
                }
            }
            $serviceStateSummary['total']++;
        }
        return $serviceStateSummary;
    }

    /**
     * @param $uuid
     * @return Servicestatus|null
     */
    public function byUuid($uuid) {
        $query = $this->find()
            ->where([
                'Servicestatus.service_description' => $uuid
            ])
            ->first();

        return $query;
    }

    /**
     * @param string|array $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byUuids($uuids, $enableHydration = false) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        if (empty($uuids)) {
            return [];
        }

        $query = $this->find()
            ->where([
                'Servicestatus.service_description IN' => $uuids
            ])
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $servicestatus = [];
        foreach ($query->toArray() as $record) {
            $servicestatus[$record['service_description']] = $record;
        }

        return $servicestatus;
    }

    /**
     * @param string|array $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byHostUuids($uuids, $enableHydration = false, $where = []) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        $where = Hash::merge([
            'Servicestatus.hostname IN' => $uuids
        ], $where);

        $query = $this->find()
            ->where($where)
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $servicestatus = [];
        foreach ($query->toArray() as $record) {
            $servicestatus[$record['service_description']] = $record;
        }

        return $servicestatus;
    }

    /**
     * @param string|array $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byUuidsWithWhere($uuids, $enableHydration = false, $where = []) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        $where = Hash::merge([
            'Servicestatus.service_description IN' => $uuids
        ], $where);

        $query = $this->find()
            ->where($where)
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $servicestatus = [];
        foreach ($query->toArray() as $record) {
            $servicestatus[$record['service_description']] = $record;
        }

        return $servicestatus;
    }

    /**
     * @param string $uuid
     * @return bool
     */
    public function existsByUuid(?string $uuid) {
        if ($uuid === null) {
            return false;
        }
        return $this->exists(['Servicestatus.service_description' => $uuid]);
    }


    /**
     * @param string|null $hostuuid
     * @return array
     */
    public function getServicesByHostUuidForAck(?string $hostuuid) {
        if ($hostuuid === null) {
            return [];
        }

        $query = $this->find()
            ->select([
                'Servicestatus.hostname',
                'Servicestatus.service_description',
                'Servicestatus.current_state'
            ])
            ->where([
                'Servicestatus.hostname'        => $hostuuid,
                'Servicestatus.current_state >' => 0
            ])
            ->disableHydration()
            ->all();

        if ($query === null) {
            return [];
        }

        return $query->toArray();
    }


    /**
     * @param string|null $hostuuid
     * @return array
     */
    public function getServicesByHostUuid(?string $hostuuid) {
        if ($hostuuid === null) {
            return [];
        }

        $query = $this->find()
            ->select([
                'Servicestatus.hostname',
                'Servicestatus.service_description',
                'Servicestatus.current_state',
                'Servicestatus.output',
                'HostMapping.display_name',
                'ServiceMapping.display_name'
            ])
            ->contain([
                'HostMapping',
                'ServiceMapping'
            ])
            ->where([
                'Servicestatus.hostname' => $hostuuid
            ])
            ->disableHydration()
            ->all();

        if ($query === null) {
            return [];
        }

        return $query->toArray();
    }

    /**
     * @param string|array $uuids
     * @param bool $enableHydration
     * @return array
     */
    public function byUuidsWithMapping($uuids, $enableHydration = false) {
        if (!is_array($uuids)) {
            $uuids = [$uuids];
        }

        if (empty($uuids)) {
            return [];
        }

        $query = $this->find()
            ->where([
                'Servicestatus.service_description IN' => $uuids
            ])
            ->contain([
                'HostMapping',
                'ServiceMapping'
            ])
            ->enableHydration($enableHydration)
            ->all();

        if (empty($query)) {
            return [];
        }

        $servicestatus = [];
        foreach ($query->toArray() as $record) {
            $servicestatus[$record['service_description']] = $record;
        }

        return $servicestatus;
    }

    /**
     * @param string $uuid
     * @param bool $enableHydration
     * @return array
     */
    public function byUuidWithMapping($uuid, $enableHydration = false) {
        $query = $this->find()
            ->where([
                'Servicestatus.service_description' => $uuid
            ])
            ->contain([
                'HostMapping',
                'ServiceMapping'
            ])
            ->enableHydration($enableHydration)
            ->first();

        if (empty($query)) {
            return [];
        }

        return $query;
    }
}
