<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace App\Model\Table;

use Acl\Model\Table\AcosTable;
use App\Filter\GenericFilter;
use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\Usergroup;
use Cake\Datasource\EntityInterface;
use Cake\ORM\Association\HasMany;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Usergroups Model
 *
 * @property UsersTable|HasMany $Users
 *
 * @method Usergroup get($primaryKey, $options = [])
 * @method Usergroup newEntity($data = null, array $options = [])
 * @method Usergroup[] newEntities(array $data, array $options = [])
 * @method Usergroup|bool save(EntityInterface $entity, $options = [])
 * @method Usergroup|bool saveOrFail(EntityInterface $entity, $options = [])
 * @method Usergroup patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Usergroup[] patchEntities($entities, array $data, array $options = [])
 * @method Usergroup findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin TimestampBehavior
 */
class UsergroupsTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('usergroups');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->addBehavior('Acl.Acl', ['requester']);

        $this->hasMany('Users', [
            'foreignKey' => 'usergroup_id'
        ]);

        $this->hasOne('Aros', [
            'className'  => 'Acl.Aros',
            'foreignKey' => 'foreign_key',
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->scalar('name')
            ->maxLength('name', 255)
            ->requirePresence('name', 'create')
            ->allowEmptyString('name', null, false);

        $validator
            ->scalar('description')
            ->maxLength('description', 255)
            ->allowEmptyString('description');

        return $validator;
    }

    /**
     * @return array
     */
    public function getUsergroupsList() {
        $query = $this->find('list');
        return $query->toArray();
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Usergroups.id' => $id]);
    }

    /**
     * @param PaginateOMat|null $PaginateOMat
     * @param GenericFilter $GenericFilter
     * @return array
     */
    public function getUsergroups($PaginateOMat, GenericFilter $GenericFilter) {
        $query = $this->find();
        $query->select([
            'Usergroups.id',
            'Usergroups.name',
            'Usergroups.description',
            'usercount' => $query->newExpr('COUNT(DISTINCT Users.id)')

        ])
            ->leftJoinWith('Users')
            ->group(['Usergroups.id'])
            ->order(array_merge(
                    $GenericFilter->getOrderForPaginator('Usergroups.name', 'asc'),
                    ['Usergroups.id' => 'asc']
                )
            )
            ->disableHydration();

        if (!empty($GenericFilter->genericFilters())) {
            $query->where($GenericFilter->genericFilters());
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }


    /**
     * @param $id
     * @return array|EntityInterface|null
     */
    public function getUsergroupById($id) {
        $query = $this->find('all')
            ->disableHydration()
            ->where([
                'id' => $id
            ]);
        if (is_null($query)) {
            return [];
        }
        return $query->first();
    }

    /**
     * @param bool $useAcoIdAsKey
     * @return array
     */
    public function getAllAcosAsList($useAcoIdAsKey = true) {
        /** @var AcosTable $AcosTable */
        $AcosTable = TableRegistry::getTableLocator()->get('Acl.Acos');

        $acosAsList = [];
        $acos = $AcosTable->find('threaded')
            ->disableHydration()
            ->all();

        $acos = $acos->toArray();

        foreach ($acos[0]['children'] as $controller) {
            if (substr($controller['alias'], -6) === 'Module') {
                $module = $controller;
                foreach ($module['children'] as $moduleController) {
                    foreach ($moduleController['children'] as $moduleAction) {
                        if ($useAcoIdAsKey === true) {
                            $acosAsList[$moduleAction['id']] = $module['alias'] . '/' . $moduleController['alias'] . '/' . $moduleAction['alias'];
                        } else {
                            $key = $module['alias'] . '/' . $moduleController['alias'] . '/' . $moduleAction['alias'];
                            $acosAsList[$key] = $moduleAction['id'];
                        }
                    }
                }
            } else {
                //Core Controller
                foreach ($controller['children'] as $action) {
                    if ($useAcoIdAsKey === true) {
                        $acosAsList[$action['id']] = $controller['alias'] . '/' . $action['alias'];
                    } else {
                        $key = $controller['alias'] . '/' . $action['alias'];
                        $acosAsList[$key] = $action['id'];
                    }
                }
            }
        }

        return $acosAsList;
    }
}
