<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class SystemLoad
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class Ntp extends BasePlugin implements PluginInterface {

    const MICROSECONDS_TO_SECONDS = 1000000;

    public function execute() {
        if ($this->checkError('ntp')) {
            return;
        }

        $requiredKeys = [
            'agent',
            'ntp'
        ];

        foreach ($requiredKeys as $requiredKey) {
            if (!isset($this->json[$requiredKey])) {
                $this->setOutput(sprintf('Key "%s" not found in Agent output', $requiredKey));
                $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                return;
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }


        $warning = '-60:60'; // warning NTP offset in seconds
        $critical = '-90:90:'; // critical NTP offset in seconds
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $warning = $arg;
                    break;
                case 1:
                    $critical = $arg;
                    break;
            }
        }

        // Test data
        //$this->json['ntp'] = [
        //    'Timestamp'      => 1708329908,
        //    'TimestampMicro' => 1708329908350812,
        //    'SyncStatus'     => true,
        //    'Offset'         => -0.000013486
        //];

        $timestamp = $this->json['ntp']['Timestamp']; // Timestamp of the last check evaluation (int)
        $timestampMicro = $this->json['ntp']['TimestampMicro']; // Timestamp of the last check evaluation in micro seconds (can be used if no NTP server is configured or on macOS) (int)
        $syncStatus = $this->json['ntp']['SyncStatus']; // Is synchronized with an NTP server (bool)
        $offset = $this->json['ntp']['Offset']; // Time offset between local system and NTP server in seconds (float)

        $output = 'Undefined state';

        if ($syncStatus === true) {
            // Server is synchronized with an NTP server
            // System clock is synchronized. Offset -0,01238712668 seconds
            $output = sprintf('System clock is synchronized. Offset %s seconds', number_format(
                $offset,
                11
            ));

        } else {
            // Remote Server is NOT synchronized
            // We use the local time of the openITCOCKPIT server for comparison

            $now = time();
            $nowMicro = $this->getMicrotimestamp();

            if ($timestamp > $now) {
                // System Clock on the client is in the future - so the time is wrong
                // Calcuate offset in micro seconds
                $offset = $timestampMicro - $nowMicro;

                // Convert Offset to Offset in Seconds
                $offset = $offset / self::MICROSECONDS_TO_SECONDS;

                $output = sprintf('System clock is in the future compared to server time. Offset %s seconds', number_format(
                    $offset,
                    11
                ));
            } else {
                // System clock on the client is behind (or eq) our clock($now)

                // Compensate for check interval of the agent and the check interval of openITCOCKPIT
                $interval = $this->json['agent']['check_interval'] ?? 30; // Check interval of the agent in seconds

                $diffFromClientToServerInSeconds = $now - $timestamp;
                if ($diffFromClientToServerInSeconds - $interval > 0) {
                    // The client clock is way behind our clock !
                    // state = warning or critical

                    // Calcuate offset in micro seconds
                    $offset = $nowMicro - $timestampMicro - ($interval * self::MICROSECONDS_TO_SECONDS);

                    // Convert Offset to Offset in Seconds
                    $offset = $offset / self::MICROSECONDS_TO_SECONDS;

                    $output = sprintf('System clock is in the behind compared to server time. Offset %s seconds', number_format(
                        $offset,
                        11
                    ));

                } else {
                    // We tolerate this kind of clock shift since we cant tell if if the clock is wrong
                    // or just of due to the check_interval of 30 seconds
                    // State = ok

                    // Calcuate offset in micro seconds
                    $offset = $nowMicro - $timestampMicro;

                    // Convert Offset to Offset in Seconds
                    $offset = $offset / self::MICROSECONDS_TO_SECONDS;

                    $output = sprintf('System clock shift is within tolerances. Offset %s seconds', number_format(
                        $offset,
                        11
                    ));

                }

            }

        }


        $this->addPerfdata(new PerfdataGauge(
            'offset',
            number_format($offset, 11, '.', ''),
            's',
            $warning,
            $critical
        ));

        $this->setOutput($output);

        if ($this->json['agent']['goos'] === 'darwin') {
            $this->setLongOutput('NTP monitoring is not supported on macOS');
        }

        if ($this->getRangeStatus($critical, $offset)) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($this->getRangeStatus($warning, $offset)) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }

        $this->setStatuscode(NagiosService::STATUS_OK);
    }

}
