<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class Process
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class Process extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('processes')) {
            return;
        }

        $requiredKeys = [
            'processes'
        ];

        foreach ($requiredKeys as $requiredKey) {
            if (!isset($this->json[$requiredKey])) {
                $this->setOutput(sprintf('Key "%s" not found in Agent output', $requiredKey));
                $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                return;
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $match = '';
        $strict = false;
        $identifier = 'auto'; // ITC-2939
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $cpuWarning = $arg;
                    break;
                case 1:
                    $cpuCritical = $arg;
                    break;
                case 2:
                    $memoryWarning = $arg;
                    break;
                case 3:
                    $memoryCritical = $arg;
                    break;
                case 4:
                    $amountWarning = $arg;
                    break;
                case 5:
                    $amountCritical = $arg;
                    break;
                case 6:
                    $match = trim($arg);
                    break;
                case 7:
                    $strict = (bool)$arg;
                    break;
                case 8:
                    $identifier = $arg;
                    break;
            }
        }

        if (!in_array($identifier, ['cmdline', 'exec', 'name', 'auto'], true)) {
            $identifier = 'auto';
        }


        $wantedProcessList = [];

        foreach ($this->json['processes'] as $i => $process) {
            if (trim($match) === '') {
                $wantedProcessList[] = $process;
                continue;
            }

            $processName = $this->getProcessName($process, $identifier);

            if ($this->processNameMatch($match, $processName, $strict)) {
                if (isset($process['ppid'])) {    //if it has a parent process
                    $parent = $this->getProcessByPID($process['ppid']);

                    if (isset($parent['ppid']) && $this->processNameMatch($match, $this->getProcessName($parent, $identifier), $strict)) {
                        $wantedProcessList[] = $this->getHighestMatchingParent($parent, $match, $strict, $identifier);
                    } else {
                        $wantedProcessList[] = $process;
                    }
                } else {    // no parent found
                    $wantedProcessList[] = $process;
                }
            }
        }

        if (empty($wantedProcessList)) {
            $this->setOutput(sprintf('No process found matching "%s"!', $match));
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        $goos = $this->json['agent']['goos'] ?? 'linux';

        $pidList = [];
        $totalMemoryPercent = 0.0;
        $totalCPUPercent = 0.0;

        // Linux and macOS
        $totalRss = 0;
        $totalVms = 0;

        // Windows
        $totalWorkingSet = 0;
        $totalWorkingSetPrivate = 0;
        $totalPrivateBytes = 0;

        // Docs https://docs.microsoft.com/en-us/windows/win32/memory/working-set
        // working_set Memory usage of the process
        // working_set_private Same value as TaskManager shows
        // private_bytes private_bytes Memory that the process use and which can not be shared with other processes

        foreach ($wantedProcessList as $process) {
            if (!in_array($process['pid'], $pidList)) {
                $pidList[] = $process['pid'];
                $totalCPUPercent += $process['cpu_percent'];
                $totalMemoryPercent += $process['memory_percent'];
                if (isset($process['Memory']['rss'])) {
                    $totalRss += $process['Memory']['rss'];
                }
                if (isset($process['Memory']['vms'])) {
                    $totalVms += $process['Memory']['vms'];
                }

                if (isset($process['Memory']['working_set'])) {
                    $totalWorkingSet += $process['Memory']['working_set'];
                }
                if (isset($process['Memory']['working_set_private'])) {
                    $totalWorkingSetPrivate += $process['Memory']['working_set_private'];
                }
                if (isset($process['Memory']['private_bytes'])) {
                    $totalPrivateBytes += $process['Memory']['private_bytes'];
                }
            }
        }

        $this->addPerfdata(new PerfdataGauge(
            'Memory_Used',
            round($totalMemoryPercent, 2),
            '%',
            $memoryWarning,
            $memoryCritical,
            0,
            100
        ));

        if ($goos === 'windows') {
            // Same value as TaskManager use
            $this->addPerfdata(new PerfdataGauge(
                'working_set_private',
                $totalWorkingSetPrivate,
                'B',
                null,
                null,
                0
            ));

            $this->addPerfdata(new PerfdataGauge(
                'working_set',
                $totalWorkingSet,
                'B',
                null,
                null,
                0
            ));

            $this->addPerfdata(new PerfdataGauge(
                'private_bytes',
                $totalPrivateBytes,
                'B',
                null,
                null,
                0
            ));
        } else {
            // Linux and macOS
            $this->addPerfdata(new PerfdataGauge(
                'rss',
                $totalRss,
                'B',
                null,
                null,
                0
            ));

            $this->addPerfdata(new PerfdataGauge(
                'vms',
                $totalVms,
                'B',
                null,
                null,
                0
            ));
        }


        $this->addPerfdata(new PerfdataGauge(
            'CPU_Used',
            round($totalCPUPercent, 2),
            '%',
            $cpuWarning,
            $cpuCritical,
            0,
            100
        ));

        $this->addPerfdata(new PerfdataGauge(
            'Amount',
            sizeof($pidList),
            '',
            $amountWarning,
            $amountCritical,
            0
        ));

        $preformat = '';
        $format = 'Amount: %s processes. CPU usage: %s%% Memory usage: %s%%';
        $this->setStatuscode(NagiosService::STATUS_OK);

        if ($this->getRangeStatus($memoryWarning, $totalMemoryPercent)) {
            $preformat = 'Memory usage warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($cpuWarning, $totalCPUPercent)) {
            $preformat = 'CPU usage warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($amountWarning, sizeof($pidList))) {
            $preformat = 'Process amount warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($memoryCritical, $totalMemoryPercent)) {
            $preformat = 'Memory usage critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }

        if ($this->getRangeStatus($cpuCritical, $totalCPUPercent)) {
            $preformat = 'CPU usage critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }

        if ($this->getRangeStatus($amountCritical, sizeof($pidList))) {
            $preformat = 'Process amount critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }


        $output = sprintf(
            $preformat . $format,
            sizeof($pidList),
            round($totalCPUPercent, 2),
            round($totalMemoryPercent, 2)
        );

        if ($goos === 'windows') {
            $memory = sprintf(
                ' Memory usage: %s',
                Byte::asHuman($totalWorkingSetPrivate)
            );
        } else {
            $memory = sprintf(
                ' RSS: %s VMS: %s',
                Byte::asHuman($totalRss),
                Byte::asHuman($totalVms)
            );
        }

        $this->setOutput($output . $memory);
    }

    private function getProcessByPID($pid) {
        foreach ($this->json['processes'] as $i => $process) {
            if (isset($process['pid']) && $process['pid'] == $pid) {
                return $process;
            }
        }
    }

    private function getHighestMatchingParent($process, $match, $strict, $identifier, $deep = 1) {
        if ($deep > 100) {
            return $process;
        }

        if (isset($process['ppid']) && $process['ppid'] > 0 && $this->processNameMatch($match, $this->getProcessName($this->getProcessByPID($process['ppid']), $identifier), $strict)) {
            $deep = $deep + 1;
            return $this->getHighestMatchingParent($this->getProcessByPID($process['ppid']), $match, $strict, $identifier, $deep);
        }
        return $process;
    }

    private function getProcessName($process, $identifier = 'auto') {
        if ($identifier !== 'auto') {
            if (isset($process[$identifier])) {
                return $process[$identifier];
            }
        }

        $processName = $process['name'];
        if (!empty($process['exec'])) {
            $processName = $process['exec'];
        }
        if (!empty($process['cmdline'])) {
            $processName = $process['cmdline'];
        }
        return $processName;
    }

    private function processNameMatch($match, $processName, $strict) {
        if ($strict && trim($match) === trim($processName)) {
            return true;
        }
        if (!$strict && strpos(trim($processName), trim($match)) !== false) {
            return true;
        }
        return false;
    }

}
