<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;

/**
 * Class SystemdService
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class SystemdService extends BasePlugin implements PluginInterface {


    public function execute() {
        if ($this->checkError('systemd_services')) {
            return;
        }

        if (!isset($this->json['systemd_services']) || !is_array($this->json['systemd_services'])) {
            $this->setOutput('Key "systemd_services" not found in Agent output');
            return;
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $unit = null;
        $strict = true;
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $unit = trim($arg);
                    break;
                case 1:
                    $strict = (bool)$arg;
                    break;
            }
        }

        foreach ($this->json['systemd_services'] as $service) {
            if ($this->serviceUnitMatch($unit, $service['Name'], $strict)) {
                $this->setOutput(sprintf('%s %s (%s)', $service['Name'], $service['LoadState'], $service['SubState']));
                $longOutput = [
                    sprintf('Name: %s', $service['Name']),
                    sprintf('Description: %s', $service['Description']),
                    sprintf('Loaded: %s', $service['LoadState']),
                    sprintf('Active: %s', $service['ActiveState']),
                    sprintf('State: %s', $service['SubState']),
                ];

                $this->setLongOutput(implode(PHP_EOL, $longOutput));

                switch ($this->getUnitType($service['Name'])) {
                    // I was not able to find a list about all available substates per unit
                    // I could only find some constants in the source code of systemd itself
                    // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c
                    case '.socket':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L219-L234
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'running' && $service['SubState'] != 'listening') {
                                // Socket is running but in a strange substate
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;

                    case '.device':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L136-L140
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'plugged') {
                                // Device is probably in "tentative" state - what ever this means
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;

                    case '.mount':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L144-L157
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'mounted') {
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;

                    case '.automount':
                    case '.path':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L127-L132
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L161-L166
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'running' && $service['SubState'] != 'waiting') {
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;


                    case '.swap':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L238-L248
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);


                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'active') {
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;

                    case '.target':
                    case '.slice':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L252-L255
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L212-L215
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] == 'active') {
                            $this->setStatuscode(NagiosService::STATUS_OK);
                        }

                        return;
                        break;


                    case '.timer':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L259-L265
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'active' && $service['SubState'] != 'waiting') {
                                // waiting == waiting for the next execution in X seconds
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }

                        return;
                        break;

                    case '.scope':
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L170-L178
                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead' && $service['SubState'] != 'failed') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'running') {
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }

                        return;
                        break;

                    default:
                        // Use .service as default
                        // https://github.com/systemd/systemd/blob/main/src/basic/unit-def.c#L182-L208

                        $this->setStatuscode(NagiosService::STATUS_CRITICAL);

                        /*
                         * States for .service units
                         * Source: https://www.computernetworkingnotes.com/linux-tutorials/systemd-units-explained-with-types-and-states.html
                         *
                         * loaded             The unit's configuration file has been successfully read and processed.
                         * Active (exited)    Successfully executed the one-time configuration and after execution, the unit is neither running an active process nor waiting for an event.
                         * Active (running)   Successfully executed the one-time configuration and after execution, the unit is running one or more active processes.
                         * Active (waiting)   Successfully executed the one-time configuration and after execution, the unit is waiting for an event.
                         * Inactive (dead)    Either the one-time configuration failed to execute or not executed yet.
                         *
                         * ----
                         * A .target unit like "nfs-client.target" can have an "active" state
                         */

                        if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead') {
                            $this->setStatuscode(NagiosService::STATUS_OK);

                            if ($service['SubState'] != 'exited' && $service['SubState'] != 'running' && $service['SubState'] != 'waiting' && $service['SubState'] != 'active') {
                                // Not sure if we will ever enter this if
                                $this->setStatuscode(NagiosService::STATUS_WARNING);
                            }
                        }
                        return;
                        break;
                }
            }
        }

        $this->setOutput(sprintf('Service (%s) not found', $unit));
        $this->setStatuscode(NagiosService::STATUS_CRITICAL);
    }

    private function serviceUnitMatch($match, $unit, $strict) {
        if ($strict && $match === $unit) {
            return true;
        }
        if (!$strict && strpos($unit, $match) !== false) {
            return true;
        }
        return false;
    }

    private function getUnitType(string $unitName): string {
        // From the Systemd unit docs (19.09.2023)
        // https://www.freedesktop.org/software/systemd/man/systemd.unit.html
        // Valid unit names consist of a "unit name prefix", and a suffix specifying the unit type which begins with a dot.
        // The "unit name prefix" must consist of one or more valid characters (ASCII letters, digits, ":", "-", "_", ".", and "\").
        // The total length of the unit name including the suffix must not exceed 255 characters. The unit type suffix must be one of
        // ".service", ".socket", ".device", ".mount", ".automount", ".swap", ".target", ".path", ".timer", ".slice", or ".scope".

        $unitTypes = ['.service', '.socket', '.device', '.mount', '.automount', '.swap', '.target', '.path', '.timer', '.slice', '.scope'];
        foreach ($unitTypes as $unitType) {
            if (strstr($unitName, $unitType)) {
                return $unitType;
            }
        }

        // No unit type matches - for now we use .service in this case
        return '.service';
    }
}
