<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class WindowsEventlog
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class WindowsEventlog extends BasePlugin implements PluginInterface {

    /**
     * @var string[]
     * https://docs.microsoft.com/de-de/dotnet/api/system.diagnostics.eventlogentrytype?view=net-5.0
     */
    private $EventLogEntryTypes = [
        0 => 'Information', // ????

        1  => 'Error',
        2  => 'Warning',
        4  => 'Information',
        8  => 'SuccessAudit',
        16 => 'FailureAudit'
    ];

    public function execute() {
        if ($this->checkError('windows_eventlog')) {
            return;
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $logType = 'Application';
        $defaultState = 0;   // (0=ok,1=warning,2=critical,3=unknown)
        $entryTypeIds = [];
        $entryTypeIdsMode = 'included';
        $eventIDs = [];
        $eventIDsMode = 'included';
        $matchSource = '';
        $matchMessage = '';

        foreach ($this->pluginConfig['args'] as $key => $arg) {
            $arg = trim($arg);
            switch ($key) {
                case 0:
                    if ($arg !== '') {
                        $logType = $arg;
                    }
                    break;
                case 1:
                    if ($arg !== '') {
                        if (is_numeric($arg) && $arg >= 0 && $arg <= 3) {
                            $defaultState = (int)$arg;
                        } else {
                            $defaultState = 3;
                        }
                    }
                    break;
                case 2:
                    if ($arg !== '') {
                        $ids = explode(',', $arg);
                        if (is_array($ids)) {
                            $entryTypeIds = $ids;
                        }
                    }
                    break;
                case 3:
                    if ($arg !== '') {
                        if ($arg !== 'included') {
                            $entryTypeIdsMode = 'excluded';
                        }
                    }
                    break;
                case 4:
                    if ($arg !== '') {
                        $ids = explode(',', $arg);
                        if (is_array($ids)) {
                            $eventIDs = $ids;
                        }
                    }
                    break;
                case 5:
                    if ($arg !== '') {
                        if ($arg !== 'included') {
                            $eventIDsMode = 'excluded';
                        }
                    }
                    break;

                case 6:
                    if ($arg !== '') {
                        $matchSource = $arg;
                    }
                    break;
                case 7:
                    if ($arg !== '') {
                        $matchMessage = $arg;
                    }
                    break;
            }
        }

        if ($matchSource !== '' && $this->isValidRegularExpression($matchSource) === false) {
            $matchSource = '';
        }
        if ($matchMessage !== '' && $this->isValidRegularExpression($matchMessage) === false) {
            $matchMessage = '';
        }

        if (!isset($this->json['windows_eventlog'][$logType])) {
            $this->setOutput(sprintf('Log type "%s" not found in Agent output', $logType));
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $entryTypeIds = $this->castToIntArray($entryTypeIds);
        $eventIDs = $this->castToIntArray($eventIDs);

        $records = $this->json['windows_eventlog'][$logType];

        if (!empty($entryTypeIds)) {
            $records = $this->filterByEntryType($records, $entryTypeIds, $entryTypeIdsMode);
        }
        if (!empty($eventIDs)) {
            $records = $this->filterByEventId($records, $eventIDs, $eventIDsMode);
        }
        if (!empty($matchSource)) {
            $records = $this->filterBySource($records, $matchSource);
        }
        if (!empty($matchMessage)) {
            $records = $this->filterByMatch($records, $matchMessage);
        }

        $EventLogEntryTypesCounter = [];
        foreach ($this->EventLogEntryTypes as $entryTypeId => $entryType) {
            $EventLogEntryTypesCounter[$entryTypeId] = 0;
        }

        foreach ($records as $record) {
            $entryTypeId = (int)$record['EntryType'];
            $EventLogEntryTypesCounter[$entryTypeId]++;
        }


        $counterString = [];
        foreach ($EventLogEntryTypesCounter as $entryTypeId => $counter) {
            if ($counter > 0) {
                $counterString[] = sprintf(
                    '%s entries of type %s',
                    $counter,
                    $this->EventLogEntryTypes[$entryTypeId]
                );
            }
        }

        foreach ($EventLogEntryTypesCounter as $entryTypeId => $counter) {
            $this->addPerfdata(new PerfdataGauge(
                strtolower($this->EventLogEntryTypes[$entryTypeId]),
                $counter,
                '',
                null,
                null,
                0
            ));
        }

        switch ($defaultState) {
            case 1:
                $defaultState = NagiosService::STATUS_CRITICAL;
                break;
            case 2:
                $defaultState = NagiosService::STATUS_WARNING;
                break;
            case 3:
                $defaultState = NagiosService::STATUS_UNKNOWN;
                break;
            default:
                $defaultState = NagiosService::STATUS_OK;
                break;
        }

        $this->setStatuscode($defaultState);
        $this->setOutput('No critical event log records found in: ' . $logType);

        if (sizeof($counterString) > 0) {
            $state = NagiosService::STATUS_OK;
            if ($EventLogEntryTypesCounter[2] > 0) {
                $state = NagiosService::STATUS_WARNING;
            }

            if ($EventLogEntryTypesCounter[1] > 0 || $EventLogEntryTypesCounter[16] > 0) {
                $state = NagiosService::STATUS_CRITICAL;
            }

            $this->setStatuscode($state);
            $output = 'Found ' . implode(', ', $counterString);
            $this->setOutput($output);
            foreach ($records as $record) {
                $this->appendLongOutput(sprintf(
                    '[%s] Time Generated: %s: Source: %s Event ID: %s Message: %s',
                    $this->EventLogEntryTypes[$record['EntryType']],
                    date('d.m.Y H:i:s', $record['TimeGenerated']),
                    $record['Source'],
                    $record['EventID'],
                    $record['Message']
                ));
            }
        }

    }

    /**
     * @param $match
     * @param $processName
     * @param $strict
     * @return bool
     */
    public function eventlogSourceNameMatch($match, $processName, $strict) {
        if ($strict && trim($match) === trim($processName)) {
            return true;
        }
        if (!$strict && strpos(trim($processName), trim($match)) !== false) {
            return true;
        }
        return false;
    }

    public function eventlogTimeMatch($timeGenerated, $checkPastMinutes) {
        $timeGeneratedTS = strtotime($timeGenerated);
        $checkPastMinutesTS = strtotime('-' . $checkPastMinutes . ' minutes');

        if ($checkPastMinutes == 0 || ($timeGeneratedTS >= $checkPastMinutesTS && $timeGeneratedTS < strtotime('now'))) {
            //event logged within the requested time range
            return true;
        }
        return false;
    }

    /**
     * @param array|int $arr
     * @return int[]
     */
    protected function castToIntArray($arr) {
        if (!is_array($arr)) {
            $arr = [$arr];
        }

        $intArr = [];
        foreach ($arr as $item) {
            $item = (int)$item;
            $intArr[$item] = $item;
        }

        return $intArr;
    }

    /**
     * @param array $records
     * @param int[] $entryTypeIds
     * @param string $entryTypeIdsMode
     * @return array
     */
    private function filterByEntryType($records, $entryTypeIds, $entryTypeIdsMode = 'included') {
        $result = [];
        foreach ($records as $record) {
            $entryTypeId = (int)$record['EntryType'];

            if ($entryTypeIdsMode === 'included') {
                if (isset($entryTypeIds[$entryTypeId])) {
                    $result[] = $record;
                }
            } else {
                if (!isset($entryTypeIds[$entryTypeId])) {
                    $result[] = $record;
                }
            }
        }
        return $result;
    }

    /**
     * @param array $records
     * @param int[] $eventIDs
     * @param string $eventIDsMode
     * @return array
     */
    private function filterByEventId($records, $eventIDs, $eventIDsMode = 'included') {
        $result = [];
        foreach ($records as $record) {
            $eventId = (int)$record['EventID'];

            if ($eventIDsMode === 'included') {
                if (isset($eventIDs[$eventId])) {
                    $result[] = $record;
                }
            } else {
                if (!isset($eventIDs[$eventId])) {
                    $result[] = $record;
                }
            }
        }
        return $result;
    }

    /**
     * @param array $records
     * @param string $regEx
     * @return array
     */
    private function filterByMatch($records, $regEx) {
        $result = [];
        foreach ($records as $record) {
            if (preg_match($regEx, $record['Message'])) {
                $result[] = $record;
            }
        }
        return $result;
    }

    /**
     * @param array $records
     * @param string $regEx
     * @return array
     */
    private function filterBySource($records, $regEx) {
        $result = [];
        foreach ($records as $record) {
            if (preg_match($regEx, $record['Source'])) {
                $result[] = $record;
            }
        }
        return $result;
    }

    /**
     * @param $regEx
     * @return bool
     */
    private function isValidRegularExpression($regEx) {
        return @preg_match($regEx, '') !== false;
    }

}
